#!/usr/local/bin/perl

# Whiplash v1.4
# Copyright (c) 1997-2000, Network Dweebs
# All Rights Reserved

use IO::Socket;
use SDBM_File;
use Fcntl;
use strict vars;
use vars qw { %CONFIG %FORM @DISALLOWED_CHANNELS %COLORCODE $DEBUG $SOCKET }; 
require ".config"; 
require "hostname.pl";

%FORM = &ReadParse;

do {
 $FORM{'ADDR'} = pack("N", $FORM{'ADDR'});
 my(@ADDR) = unpack("C4", $FORM{'ADDR'});
 $FORM{'ADDR'} = join(".", @ADDR);
};

$CONFIG{'CONNECTED'}=0;
$DEBUG=0;

# Enter our PID and SHMKEY into the PID Table for Leak Cleanup
# Some operating systems do not report any signal upon losing
# the client, hence cleanup never gets called.
do {
 my(%PIDTABLE);
 tie %PIDTABLE, SDBM_File, $CONFIG{'PIDTABLE'}, O_RDWR | O_CREAT, 0644
   || die "tie: $!\n";
 $CONFIG{'PIDTABLE'}{$FORM{'IPCKEY'}}=$$;
 untie $CONFIG{'PIDTABLE'};
};


$CONFIG{'IPC_RMID'} = 0;

# We use a : separator here and a \002 separator in the message because
# When you overwrite shared memory, it only overwrites the length of
# the message, leaving any old junk lying around.

&readymem;

sub readymem {
 my($MESSAGE) = "SOCKET:$$:";
 &pshmwrite($FORM{'IPCKEY'}, $MESSAGE, 0, length($MESSAGE));
}

# Here we set the various methods of scrolling,
# based on what browser version the client has.
#
# Netscape v3.0 does not support scrollBy, and as a result
# is flaky its scrolling (since we have to scroll to line
# 1000000, which usually doesn't exist. 

if ($FORM{'BROWSER'} eq "NETSCAPE") {
 if ($FORM{'BROWSER_VERSION'}<4) { $CONFIG{'SCROLL'}="scroll(1,10000000)"; } 
 else { $CONFIG{'SCROLL'}="self.scrollBy(0,100)"; }
}
else { $CONFIG{'SCROLL'}="self.scrollBy(0,100)"; }

if ($ENV{'HTTP_USER_AGENT'} =~ /Macintosh|PowerPC/i) {
 $CONFIG{'SCROLL'} = "self.scrollBy(0,13);";
}

setpriority 0, 0, getpriority(0, 0) + $CONFIG{'NICE'};

$| = 1;

$SIG{'HUP'}  = \&dieout;
$SIG{'INT'}  = \&dieout;
$SIG{'ALRM'} = \&msg;

sub dieout {
 my($SIG)=@_;
 my(%PIDTABLE);
 if ($CONFIG{'USE_SHMEM'}==1)
  { shmctl($FORM{'IPCKEY'} ,$CONFIG{'IPC_RMID'}, 0); }
 else { unlink("$CONFIG{'MEMDIR'}/$FORM{'IPCKEY'}"); }
 tie %PIDTABLE, SDBM_File, $CONFIG{'PIDTABLE'}, O_RDWR | O_CREAT, 0644;
 delete($CONFIG{'PIDTABLE'}{$FORM{'IPCKEY'}});
 untie $CONFIG{'PIDTABLE'};
 close $SOCKET;
 exit;
}

# Process the incoming message or command (Triggered by an ALRM)
sub msg {
 my($MEMORY)=&pshmread($FORM{'IPCKEY'}, 0, 512);
 ($MEMORY)=split(/\002/, $MEMORY);
 chomp($MEMORY);
 my($COMMAND, @LINE) = split(/ /, $MEMORY);
 my($LINE) = join(" ", @LINE);
 if ($COMMAND !~ /^\//) { $LINE = $COMMAND . " " . $LINE; $COMMAND=""; }
 $COMMAND =~ tr/a-z/A-Z/;

 if ($COMMAND eq "/QUIT") { 
  close $SOCKET;
  &output("<FONT COLOR=#FF0000><B>DCC Connection Closed.</B></FONT>");
  &dieout;  
 }

 if ($COMMAND eq "/CLEAR") { 
  if ($FORM{'BROWSER'} eq "NETSCAPE") {
   print "--BOUNDARY\n";
   print "Content-type: text/html\n\n";
   print "<HTML><BODY BGCOLOR=#FFFFFF TEXT=#000000>";
   &header;
  }
  &readymem; 
  return;
 }

 if ($COMMAND eq "/ME") { 
  $SOCKET->send("\001ACTION $LINE\001\015\012");
  &output("* $FORM{'NICK'} $LINE"); 
  &readymem;
  return;
 }

 if ($COMMAND eq "") {
  &output("&lt;$FORM{'NICK'}&gt; $LINE");
  $SOCKET->send("$LINE\015\012") || &dieout;
 }
 else { &output("Unknown Command: $COMMAND"); }
 &readymem;
 return;
}

# Print the magic header to create a server push, also
# referred to as a 'server stream'.  Great for chatrooms *g*
 
print "HTTP/1.0 200\n";
print "Content-type: multipart/x-mixed-replace;boundary=BOUNDARY\n\n";
if ($FORM{'BROWSER'} eq "MSIE") {
 print "<HTML><BODY BGCOLOR=#FFFFFF TEXT=#000000>"; }
elsif ($FORM{'BROWSER'} eq "NETSCAPE") {
 print "--BOUNDARY\n";
 print "Content-type: text/html\n\n";
 print "<HTML><BODY BGCOLOR=#FFFFFF TEXT=#000000>";
}
print "<CODE>\n" if ($FORM{'fixed'});


$0="DCC CHAT: $FORM{'NICK'}";

&header;
sub header {
 &output("<FONT COLOR=#FF0000><B><I>Whiplash v$CONFIG{'VERSION'}</i></B>" .
         " </FONT><FONT COLOR=#0000FF>- (c) 1997-2000, Network Dweebs");
 &output("<B>DCC Chat with $FORM{'THEIRNICK'}</B>");
 &output("<HR></FONT>");

 print "<STYLE TYPE=\"text/css\">\n<!--\n";
 foreach(keys(%COLORCODE)) {
  print ".C$_ {background-color:$COLORCODE{$_};}\n";
 }
 print "//-->\n</STYLE>\n";

 return;
}

do {
 my($x)=&buildsocket($FORM{'ADDR'}, $FORM{'PORT'});
 if ($x != 1) { &output($x); &dieout }
};

while(1) {
 my($SOCKETIN);
 $SIG{'ALRM'} = \&msg;
 $SOCKET->recv($SOCKETIN, 10240);

 if (length($SOCKETIN)<1) { 
  &output("<FONT COLOR=#FF0000><B>DCC Connection Closed.</B></FONT>"); 
  &dieout;
 }

 chomp($SOCKETIN);
  
 if ($SOCKETIN =~ /^\001ACTION /) {
  $SOCKETIN =~ s/^\001ACTION //;
  $SOCKETIN =~ s/\001$//;
   &output("<FONT COLOR=#FF00FF>* $FORM{'THEIRNICK'} $SOCKETIN</FONT>");
 }
 else { &output("&lt;$FORM{'THEIRNICK'}&gt; $SOCKETIN"); }

}

sub colorcode {
 my($MESSAGE)=@_;
 return $MESSAGE unless (/\|\|\|\|\/);
 my($REV, $BOLD, $UL, $COL, $POSITION)=0;
 my($NEWMESSAGE)="";
 $MESSAGE .= "";
 while($POSITION<=length($MESSAGE)) {
  my($CHAR)=substr($MESSAGE, $POSITION, 1);
  if (($CHAR eq "") && ($BOLD == 0)) { $CHAR = "<B>"; $BOLD=1; }
  if (($CHAR eq "") && ($BOLD == 1)) { $CHAR = "</B>"; $BOLD=0; }
  if (($CHAR eq "") && ($UL == 0)) { $CHAR = "<U>"; $UL=1; }
  if (($CHAR eq "") && ($UL == 1)) { $CHAR = "</U>"; $UL=0; }
  if ($FORM{'BROWSER_VERSION'}>=4) {
   if (($CHAR eq "") && ($REV == 0)) { $CHAR = "<SPAN CLASS=C1>" .
     "<FONT COLOR=#FFFFFF>"; $REV=1; }
   if (($CHAR eq "") && ($REV == 1)) { $CHAR = "</FONT></SPAN>"; $REV=0; }
  }
  if ($CHAR eq "") {
   my($CODE, $FG, $BG)="";
   $CODE = "</FONT></SPAN>" if ($COL==1);
   $CHAR=substr($MESSAGE, $POSITION, 6);
   if ($CHAR =~ /^\d+(\,\d+)?/) {
    my($match) = $&;
    $match =~ s/^\//;
    ($FG, $BG)=split(/\,/, $match);
    $COL=1;  
    $POSITION += length($match);
   } 
   else { $COL=0; }
   if ($FG ne "") { 
    if ($BG < 1 ) { $BG = "0"; }
    $CODE .= "<SPAN CLASS=C$BG><FONT COLOR=$COLORCODE{$FG}>"; 
   }
   $CHAR=$CODE;
  }
  if ($CHAR eq "") {
   my($CODE)="";
   $CODE .= "</B>" if ($BOLD==1);
   $CODE .= "</U>" if ($UL==1);
   $CODE .= "</FONT></SPAN>" if ($COL==1);
   $CODE .= "</FONT></SPAN>" if ($REV==1);
   $BOLD=0; $UL=0; $COL=0; $REV=0;
   $CHAR=$CODE;
  }
  $NEWMESSAGE .= $CHAR;
  $POSITION++;
 }
 return $NEWMESSAGE;
} 

# Builds our tcp socket with the irc server

sub buildsocket {
 my($HOST, $PORT) = @_;
 if (!$PORT) { my($x)=&buildlisteningsocket; return $x; }
 else {
  &output("<B><FONT COLOR=#FF0000>Acknowleging DCC Request...</FONT></B>");
 }

 $SOCKET = new IO::Socket::INET( Proto    => "tcp",
			  PeerAddr => "$HOST:$PORT" );

 if (defined $SOCKET) {
  &output("<FONT COLOR=#FF0000><B>DCC Connection Established</B></FONT>");
 }
 else {
  &output("Unable to connect to client: $!");     
  close $SOCKET;
  &dieout;
 }

 return 1;
}

sub buildlisteningsocket {
 my($HOSTNAME) = &hostname;

 $SOCKET = new IO::Socket::INET(Listen    => 1,
               LocalAddr => $HOSTNAME,
               LocalPort => &Socket::INADDR_ANY(),
               Proto     => 'tcp');

 my($PORT) = $SOCKET->sockport();
 &output("<FONT COLOR=#FF0000><B>Waiting for DCC Connection...</B></FONT>");
 &output("Listening on port $PORT...");

 $SOCKET->autoflush(1);
 $SOCKET->timeout($CONFIG{'DCC_TIMEOUT'});

 my($buff)=&pshmread($FORM{'SOCKETKEY'}, 0, 20);
 my($wait)=0;
 my($pid);
 while($buff !~ /^SOCKET\:/) {
  $wait++;
  $buff=&pshmread($FORM{'SOCKETKEY'}, 0, 20);
  exit if ($wait>25);
  select(undef, undef, undef, .50);
 }
 ($pid, $pid)=split(/\:/, $buff);

 my($name, $aliases, $type, $len, $ip)=gethostbyname($HOSTNAME);
 my(@ip)=unpack("C4", $ip);
 my($IP) = unpack ("N", pack("C4", @ip, 2));

 my($message) = "/CTCP $FORM{'THEIRNICK'} DCC CHAT chat $IP $PORT\002";
 &pshmwrite($FORM{'SOCKETKEY'}, $message, 0, length($message));
 kill ALRM, $pid;

 $SOCKET = $SOCKET->accept;

 if (!defined $SOCKET) { 
  &output("<FONT COLOR=#FF0000><B>DCC Connection Timed out.</B></FONT>"); 
  &dieout; 
 }

 &output("<B><FONT COLOR=#FF0000>DCC Session Established</FONT></B>");

 return 1;
}

sub output {
 my($output)=@_;
 chomp($output);
 $output = &colorcode($output);
 print "$output<BR><SCRIPT>$CONFIG{'SCROLL'}</SCRIPT>\015\012";
 return;
}


sub ReadParse {
 my(%F, $buffer, @pairs);
 if ($ENV{'REQUEST_METHOD'} =~ /GET/i)
  { @pairs = split(/&/, $ENV{'QUERY_STRING'}); }
 else {
  read(STDIN, $buffer, $ENV{'CONTENT_LENGTH'});
  @pairs = split(/\&/, $buffer);
 }
 foreach(@pairs) {
  my($name, $value) = split(/\=/, $_);

  $name =~ tr/+/ /;
  $name =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
  $value =~ tr/+/ /;
  $value =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
  $value =~ s/<!--(.|\n)*-->//g;
  $F{$name}=$value;
 }

 return %F;
}

# We need a compare subroutine because using a regex
# could result in an inaccurate result, since people
# can have |, ^, and other special characters in
# their handle.

sub compare {
 my($a, $b) = @_;
 $a =~ tr/A-Z/a-z/;
 $b =~ tr/A-Z/a-z/;
 chomp($a, $b);
 $a =~ s/^\s+//; $a =~ s/\s+$//;
 $b =~ s/^\s+//; $b =~ s/\s+$//;
 if ($a eq $b) { return 1; }
 return 0;
}

sub pshmwrite {
 my($KEY, $MESSAGE, $ID, $LENGTH)=@_;
 if ($CONFIG{'USE_SHMEM'}==1) {
  shmwrite($KEY, $MESSAGE, $ID, $LENGTH) || die "shmwrite: $!\n"; }
 else {
  open(MEM, ">$CONFIG{'MEMDIR'}/$KEY")
   || die "$CONFIG{'MEMDIR'}/$KEY: $!\n";
  print MEM $MESSAGE, "\n";
  close(MEM);
 }
 return;
}

sub pshmread {
 my($KEY, $ID, $LENGTH)=@_;
 my($BUFFER);
 if ($CONFIG{'USE_SHMEM'}==1) 
  { shmread($KEY, $BUFFER, $ID, $LENGTH) || die "shmread: $!\n"; }
 else {
  open(MEM, "<$CONFIG{'MEMDIR'}/$KEY") || die "$CONFIG{'MEMDIR'}/$KEY: $!\n";
  chomp; $BUFFER = <MEM>;
  close(MEM);
 }
 return $BUFFER;
}


sub pshmwrite {
 my($KEY, $MESSAGE, $ID, $LENGTH)=@_;
 if ($CONFIG{'USE_SHMEM'}==1) {
  shmwrite($KEY, $MESSAGE, $ID, $LENGTH) || die "shmwrite: $!\n"; }
 else {
  open(MEM, ">$CONFIG{'MEMDIR'}/$KEY") || die "$CONFIG{'MEMDIR'}/$KEY: $!\n";
  print MEM $MESSAGE, "\n";
  close(MEM);
  }
 return;
}

sub pshmread {
 my($KEY, $ID, $LENGTH)=@_;
 my($BUFFER);
 if ($CONFIG{'USE_SHMEM'}==1) 
  { shmread($KEY, $BUFFER, $ID, $LENGTH) || die "shmread: $!\n"; }
 else {
  open(MEM, "<$CONFIG{'MEMDIR'}/$KEY") || die "$CONFIG{'MEMDIR'}/$KEY: $!\n";
  chomp; $BUFFER = <MEM>;
  close(MEM);
 }
 return $BUFFER;
}

