#!/usr/local/bin/perl

# Whiplash v1.4
# Copyright (c) 1997-2000, Network Dweebs
# All Rights Reserved

use Socket;
use SDBM_File;
use Fcntl;
use IO::Socket;
use strict vars;
use vars qw { %CONFIG @DISALLOWED_CHANNELS %COLORCODE %FORM $SOCKET 
              %parse %outcome @user @phrase $DEBUG %CURRENT_USERS
              %CURRENT_MODES $LASTKNOWNPID };
require ".config"; 
require "hostname.pl";

%FORM = &ReadParse;
my($KEY) = $FORM{'IPCKEY'};

$CONFIG{'CONNECTED'}=0;
$DEBUG=0;

# Enter our PID and SHMKEY into the PID Table for Leak Cleanup
# Some operating systems do not report any signal upon losing
# the client, hence cleanup never gets called.
do {
 my(%PIDTABLE);
 tie %PIDTABLE, SDBM_File, $CONFIG{'PIDTABLE'}, O_RDWR | O_CREAT, 0644
  || die "tie: $!\n";
 $PIDTABLE{$KEY}=$$;
 untie %PIDTABLE;
};

$CONFIG{'IPC_RMID'} = 0;

# We use a : separator here and a \002 separator in the message because
# When you overwrite shared memory, it only overwrites the length of
# the message, leaving any old junk lying around.

&readymem;

sub readymem {
 my($MESSAGE) = "SOCKET:$$:";
 &pshmwrite($KEY, $MESSAGE, 0, length($MESSAGE));
}

# Here we set the various methods of scrolling,
# based on what browser version the client has.
#
# Netscape v3.0 does not support scrollBy, and as a result
# is flaky its scrolling (since we have to scroll to line
# 1000000, which usually doesn't exist. 

if ($FORM{'BROWSER'} eq "NETSCAPE") {
 if ($FORM{'BROWSER_VERSION'}<4) { $CONFIG{'SCROLL'}="scroll(1,10000000)"; } 
 else { $CONFIG{'SCROLL'}="self.scrollBy(0,100)"; }
}
else { $CONFIG{'SCROLL'}="self.scrollBy(0,100)"; }

if ($ENV{'HTTP_USER_AGENT'} =~ /Macintosh|PowerPC/i) {
 $CONFIG{'SCROLL'} = "self.scrollBy(0,13);";
}

setpriority 0, 0, getpriority(0, 0) + $CONFIG{'PRIORITY'};

# Here we load in the rfc1459 ruleset and turn it into a hash database.
open(FILE, "<.rfc1459");
while(<FILE>) {
 my($key, $output, $second) = split(/\|/);
 $parse{$key}=$output;
 $outcome{$key}=$second;
}
close(FILE);

$| = 1;

$SIG{'HUP'}  = \&dieout;
$SIG{'INT'}  = \&dieout;
$SIG{'ALRM'} = \&msg;

sub dieout {
 ($SIG)=@_;
 my(%PIDTABLE);
 if ($CONFIG{'USE_SHMEM'}==1) { shmctl($KEY ,$CONFIG{'IPC_RMID'}, 0); }
 else { unlink("$CONFIG{'MEMDIR'}/$KEY"); }
 tie %PIDTABLE, SDBM_File, $CONFIG{'PIDTABLE'}, O_RDWR | O_CREAT, 0644;
 delete($PIDTABLE{$KEY});
 untie %PIDTABLE;
 close($SOCKET);
 exit;
}

# Process the incoming message or command (Triggered by an ALRM)
sub msg {
 my($NOPARSE)=0;
 my($outcome, $output, $command);
 my($line)=&pshmread($KEY, 0, 512);
 ($line)=split(/\002/, $line);
 chomp($line);

 # Send it as a  message unless it begins with a /
 if ($line !~ /^\//) { print $SOCKET "PRIVMSG $FORM{'channel'} :$line\015\012"; }
 else {
  chomp($line);
  (@phrase)=split(/ /, $line);
  $phrase[0] =~ tr/a-z/A-Z/;
  if ($phrase[0] eq "/CLEAR") {
   if ($FORM{'BROWSER'} eq "NETSCAPE") {
    print "--BOUNDARY\n";
    print "Content-type: text/html\n\n";
    print "<HTML><BODY BGCOLOR=#FFFFFF TEXT=#000000>";
    &header;
   }
   &readymem;
   return;
  }

  if (($phrase[0] eq "/JOIN") && ($CONFIG{'ALLOWJOIN'} == 1)) {
   foreach(@DISALLOWED_CHANNELS) { 
    if ($phrase[1] =~ /$_/i) {
     &output("4*** Sorry you cannot join $phrase[1]");
     &readymem; return;
    }
   }

   print $SOCKET "PART $FORM{'channel'}\015\012"; 
   $FORM{'channel'}=$phrase[1]; 
   print $SOCKET "JOIN $FORM{'channel'}\015\012"; 
   &output("4[Joining Channel: $FORM{'channel'}]");

   &readymem;
   return;
  }

  if (($phrase[0] eq "/DCC") && ($phrase[1] =~ /chat/i)) {
   my($ADDR, $PORT)=0;
   my($URL) = "?fixed=$FORM{'fixed'}" .
              "&BROWSER=$FORM{'BROWSER'}" .
              "&BROWSER_VERSION=$FORM{'BROWSER_VERSION'}" .
              "&NICK=$FORM{'nick'}" .
              "&THEIRNICK=$phrase[2]" .
              "&ADDR=$ADDR" .
              "&SOCKETKEY=$KEY" .
              "&PORT=$PORT";
   print <<_END;
<SCRIPT>window.open('dcc-login.cgi$URL', '_dcc$phrase[2]')</SCRIPT>
_END
   &readymem;
   return;
  }

  if ($parse{$phrase[0]} eq "") {
   &output("4* Unknown Command '$phrase[0]'");
  }
  else {
   ($command)=shift(@phrase);
   ($output)=$parse{$command};
   ($outcome)=$outcome{$command};
          
   # Replace Tokens with their subsequent values
   # %2 = 2nd parameter, %2% = 2nd parameter -> end, etc.
   my($num);
   for $num (2..9) {
    my($parms)="";
    for($num..$#phrase+2) 
     { $parms .= $phrase[$_ - 2] . " "; }
    $phrase[$num - 2] =~ s/</\&lt\;/g;
    $phrase[$num - 2] =~ s/>/\&gt\;/g;
    $parms =~ s/</\&lt\;/g;
    $parms =~ s/>/\&gt\;/g; 
    $parms =~ s/\s+$//;

    $output =~ s/\%$num\%/$parms/g;
    $output =~ s/\%$num/$phrase[$num - 2]/g;
    $outcome =~ s/\%$num\%/$parms/g;
    $outcome =~ s/\%$num/$phrase[$num - 2]/g;
   }

   $output =~ s/\%CHAN/$FORM{'channel'}/g;
   $output =~ s/\%NICK/$FORM{'nick'}/g;
   $outcome =~ s/\%NICK/$FORM{'nick'}/g;
   $outcome =~ s/\%CHAN/$FORM{'channel'}/g;
   if ($phrase[0] eq "") { $output =~ s/\://g; }
   chomp($output); 
   $output =~ s/\s+$//; 
   print $SOCKET "$output\015\012";
   chomp($outcome); $NOPARSE=1;
  }
 }
 chomp($line);
 $line =~ s/</\&lt\;/g; 
 $line =~ s/>/\&gt\;/g;
 if ($line !~ /^\//) { &output("&gt; $line\n"); }
 if (($line =~ /^\//) && ($outcome ne "")) {
  $outcome =~ s/</\&lt\;/g unless ($NOPARSE);
  $outcome =~ s/>/\&gt\;/g unless ($NOPARSE);
  &output($outcome); 
 }

 if (($command eq "/NICK") && ($CONFIG{'CONNECTED'} !=1)) 
  { $FORM{'nick'}=$phrase[0]; }

 &readymem;
 return;
}

# Print the magic header to create a server push, also
# referred to as a 'server stream'.  Great for chatrooms *g*
 
print "HTTP/1.0 200\n";
print "Content-type: multipart/x-mixed-replace;boundary=BOUNDARY\n\n";
if ($FORM{'BROWSER'} eq "MSIE") {
 print "<HTML><BODY BGCOLOR=#FFFFFF TEXT=#000000>"; }
elsif ($FORM{'BROWSER'} eq "NETSCAPE") {
 print "--BOUNDARY\n";
 print "Content-type: text/html\n\n";
 print "<HTML><BODY BGCOLOR=#FFFFFF TEXT=#000000>";
}
print "<CODE>\n" if ($FORM{'fixed'});

$FORM{'nick'}=$CONFIG{'DEFAULTNICK'} unless ($FORM{'nick'});
$FORM{'server'}=$CONFIG{'DEFAULTSEVER'} unless ($FORM{'server'});
$FORM{'port'}=$CONFIG{'DEFAULTPORT'} unless ($FORM{'port'});
$FORM{'realname'}="Unknown" unless ($FORM{'realname'});

do {
 my($LONGIP) = unpack ("N", pack("C4", split(/\./, $ENV{'REMOTE_ADDR'}), 2));
 $FORM{'realname'} .= " ($LONGIP)";
};

if ($CONFIG{'DEDICATEDSEVER'} == 1) { 
 $FORM{'channel'}=$CONFIG{'DEFAULTCHANNEL'};
 $FORM{'server'}=$CONFIG{'DEFAULTSEVER'};
 $FORM{'port'}=$CONFIG{'DEFAULTPORT'} if ($CONFIG{'DEFAULTPORT'});
}

$0="IRC: $FORM{'nick'}";

# Please keep this intact.  I would like credit for what I've
# Programmed.  It is also in the license.  Thanks.

&header;
sub header {
 &output("4<I>Whiplash v$CONFIG{'VERSION'}</i>" .
         " 12- (c) 1997-2000, Network Dweebs");
 &output("12http://www.networkdweebs.com/");
 &output("12Copyright (c) 1997-2000, All Rights Reserved");
 &output("<HR>");

 print "<STYLE TYPE=\"text/css\">\n<!--\n";
 foreach(keys(%COLORCODE)) {
   print ".C$_ {background-color:$COLORCODE{$_};}\n";
 }
 print "//-->\n</STYLE>\n";

 return;
}

&output("Auto-Configured for $FORM{'BROWSER'} v$FORM{'BROWSER_VERSION'}.");
&output("Connecting to $FORM{'server'}/$FORM{'port'}...");
&output();
sleep 5;
my($x) = &buildsocket($FORM{'server'}, $FORM{'port'}, $FORM{'nick'}, 
                      $FORM{'realname'});
if ($x != 1) { &output($x); &dieout }

while(<$SOCKET>) {
 my($socketin)=$_;
 my($from, $code, $to, $msg) = &parse($_);
 my($MESSAGE);
 my($han, $addy) = &from($from);
 $han =~ s/
|\013|\n//g;
 $addy =~ s/
|\013|\n//g;
 $msg =~ s/</\&lt\;/g;
 $msg =~ s/>/\&gt\;/g;

 my($FOUND)=0;
 if ($parse{$code} ne "") {
  chomp($to);
  my($output)=$parse{$code};
  my(@parms)=split(/ /, $msg);  
  for(0..$#parms) { $parms[$_] =~ s/^\://g; }
  my($num);
  for $num (1..10) {
   my($parmline);
   for($num..$#parms+1)
    { $parmline .= $parms[$_ - 1] . " "; }
   $output =~ s/\%$num\%/$parmline/g;
   $output =~ s/\%$num/$parms[$num - 1]/g;
  }
  $output =~ s/\%TO/$to/g;
  $output =~ s/\%CHAN/$FORM{'channel'}/g;
  $output =~ s/\%HANDLE/$han/g;
  $output =~ s/\%ADDRESS/$addy/g;
  &output($output); $FOUND=1;
 } 

 # Server PING/PONG Replies (Very Important)
 if (/^PING /) {
  my($junk, $reply, $secondary) = split(/ /);
  if ($secondary eq "") { print $SOCKET "PONG $reply\015\012"; }
  else { print $SOCKET "PONG $reply $secondary\015\012"; }
  if ($CONFIG{'SHOWSPONG'} == 1)
   { &output("4* Ping? PONG!"); }
  next; 
 }

### PUT SPECIAL PARSING CODE HERE THAT WILL ALSO HAVE AN RFC ENTRY

 # CHANGING HANDLES REQUIRES $FORM{'nick'} TO BE UPDATED
 if ($code eq "NICK") {
  if (&compare($FORM{'nick'}, $han)) { $FORM{'nick'}=$to; }
 }

 # SIMULATED JOIN CHANNEL
 if ($code eq "376") { 
  $CONFIG{'CONNECTED'}=1; 
  foreach(@DISALLOWED_CHANNELS) {
   if ($phrase[1] =~ /$_/i) {
    &output("4*** Sorry you cannot join $phrase[1]");
    next;
   }
  }

  &output("Joining $FORM{'channel'}..."); print $SOCKET "JOIN $FORM{'channel'}\015\012"; next; }

  # NAMES LIST COLLECTION (For Userlist)
  if ($code eq "353") {
   my(@fields)=split(/\ /, $msg);
   shift(@fields); shift(@fields);
   $fields[0] =~ s/^\://;
   for(0..$#fields) { 
    push(@user, $fields[$_]) unless ($fields[$_] eq "");
   }
  }
 
  if ($code eq "366") {
   my($USERLIST)=join(" ", @user);
   undef(@user);
   my($MESSAGE) = "\004NEW $USERLIST\002";
   &userlist($MESSAGE);
  }


  # We need to update the userlist if any of the following actions occur...
  if ($code eq "JOIN") {
   &userlist("\004ADD $han\002") unless (&compare($han, $FORM{'nick'}));}

  if ($code eq "PART") {
   &userlist("\004REMOVE $han\002") unless (&compare($han, $FORM{'nick'})); }

  if ($code eq "QUIT") {
   &userlist("\004REMOVE $han\002"); }

  if ($code eq "MODE") {
   my($mode, @han)=split(/\ /, $msg);
   if ($mode =~ /\+v|\+o/i) {
    my(@list);
    foreach(@han) {
     next unless($_);
     if ($mode =~ /\+v/i) { push(@list, "+$_"); }
     if ($mode =~ /\+o/i) { push(@list, "\@$_"); }
    }
    $MESSAGE = "\004ADD " . join(" ", @list). "\002";
    &userlist($MESSAGE);
   }
   if ($mode =~ /\-v|\-o/i) {
    $MESSAGE="\004ADD ". join(" ", @han) . "\002";
    &userlist($MESSAGE);
   }
  }

  if ($code eq "KICK") {
   my($han)=split(/\ /, $msg);
   &userlist("\004REMOVE $han\002");
  }

  if ($code eq "NICK") { &userlist("\004NICK $han:$to\002"); }
    
### END SPECIAL PARSING CODE

  next if ($FOUND==1);
  
  # ERRORS
  if ($from eq "ERROR") { &output("$code $to $msg"); next; } 
  if ($from eq "NOTICE") { &output("$to $msg"); next; }
  
  # STANDARD MESSAGES 
  if (($code eq "NOTICE") || ($code eq "PRIVMSG")) {
   my($from, $x, $to, @msg) = split(/ /);
   my($person, $host) = split(/\!/, $from);
   $person =~ s/^\://;
   $msg =~ s/^\://;
   $msg =~ s/^\s+//g; $msg =~ s/^\t+//g; 

   # CTCP REPLIES
   if ($msg =~ /^\001PING /) {
    $msg =~ s/^\001PING //;
    $msg =~ s/\001$//;
    &output("4[PING: $person]");
    print $SOCKET "NOTICE $person :\001PING $msg\001\015\012";
    next;
   }

   if ($msg =~/^\001VERSION\001/) {
    &output("4[VERSION: $person]");
    print $SOCKET "NOTICE $person :\001VERSION Whiplash $CONFIG{'VERSION'} " .
            "www.networkdweebs.com\001\015\012";
    next;
   }
 
   if ($msg =~ /^\001USERINFO\001/) {
    &output("4[USERINFO: $person]");
    print $SOCKET "NOTICE $person :\001USERINFO $FORM{'nick'} is ".
     "$FORM{'realname'}, $ENV{'REMOTE_ADDR'}\001\015\012";
    next;
   }

   if ($msg =~ /^\001CLIENTINFO\001/) {
    &output("4[CLIENTINFO: $person]");
    print $SOCKET "NOTICE $person :\001CLIENTINFO " .
                  "PING USERINFO VERSION CLIENTINFO\001\015\012";
    next;
   }

   if ($msg =~ /^\001DCC CHAT chat /i) {
    my($line)=$msg;
    $line =~ s/^\001DCC CHAT chat //i; 
    $line =~ s/\001$//;
    chomp($line);
    my($ADDR, $PORT)=split(/ /, $line);
    my($URL) = "?fixed=$FORM{'fixed'}" .
                 "&BROWSER=$FORM{'BROWSER'}" .
                 "&BROWSER_VERSION=$FORM{'BROWSER_VERSION'}" .
                 "&NICK=$FORM{'nick'}" .
                 "&THEIRNICK=$han" .
                 "&ADDR=$ADDR" .
                 "&SOCKETKEY=$KEY" .
                 "&PORT=$PORT";
    print <<_END;
<SCRIPT>if (window.confirm("Accept DCC Chat from $han?")) { window.open('dcc-login.cgi$URL', '_dcc$han'); } </SCRIPT>
_END
    next;
   }
   ## END CTCP REPLIES

   if ($msg =~ /^\001ACTION /) {
    $msg =~ s/^\001ACTION //;
    $msg =~ s/\$//g;
    &output("13* $person $msg");
    next; }
   else {
    if ($to eq "AUTH") { &output("4$msg"); }
    elsif (&compare($to, $FORM{'nick'})) {
     if ($msg =~ /^\001/) { 
      $msg =~ s/^|\$//g;
      &output("4[CTCP: $han] $msg");
     }
     else { &output("12**$person** $msg"); } 
    } 
    else { &output("&lt;$person&gt; $msg"); } 
    next;
   }
  }
  chomp($socketin); &output("UNKNOWN: $socketin"); next;
}
&output("<BR>4Connection to host lost");
&output("4You are now disconnected from the server.");
&output("Whiplash v$CONFIG{'VERSION'} (c) 1997-2000, Network Dweebs");
&output("http://www.networkdweebs.com");
&dieout;

sub parse {
  my($in)=@_;
  my($from, $code, $to, @msg)=split(/ /, $in);
  $from =~ s/^\://; $to =~ s/^\://;
  my($msg) = join(" ", @msg);
  $from =~ s/^\://; $to =~ s/^\://; $msg =~ s/^\://;
  $msg =~ s/
|\013|\n//g;
  $from =~ s/
|\013|\n//g;
  $to =~ s/
|\013|\n//g; 
  $code =~ s/
|\013|\n//g;
  return ($from, $code, $to, $msg);
}

sub colorcode {
  my($MESSAGE)=@_;
  return $MESSAGE unless ($MESSAGE =~ /\|\|\|\|\/);
  my($REV, $BOLD, $UL, $COL, $POSITION)=0;
  my($NEWMESSAGE)="";
  $MESSAGE .= "";
  while($POSITION<=length($MESSAGE)) {
    my($CHAR)=substr($MESSAGE, $POSITION, 1);
    if (($CHAR eq "") && ($BOLD == 0)) { $CHAR = "<B>"; $BOLD=1; } 
    if (($CHAR eq "") && ($BOLD == 1)) { $CHAR = "</B>"; $BOLD=0; }
    if (($CHAR eq "") && ($UL == 0)) { $CHAR = "<U>"; $UL=1; }
    if (($CHAR eq "") && ($UL == 1)) { $CHAR = "</U>"; $UL=0; }
    if ($FORM{'BROWSER_VERSION'}>=4) {
      if (($CHAR eq "") && ($REV == 0)) { $CHAR = "<SPAN CLASS=C1>" .
        "<FONT COLOR=#FFFFFF>"; $REV=1; }
      if (($CHAR eq "") && ($REV == 1)) { $CHAR = "</FONT></SPAN>"; $REV=0; }
    }
    if ($CHAR eq "") {
      my($CODE, $FG, $BG)="";
      $CODE = "</FONT></SPAN>" if ($COL==1);
      $CHAR=substr($MESSAGE, $POSITION, 6);
      if ($CHAR =~ /^\d+(\,\d+)?/) {
        my($match) = $&;
        $match =~ s/^\//;
        ($FG, $BG)=split(/\,/, $match);
        $COL=1;  
        $POSITION += length($match);
      } 
      else { $COL=0; }
      if ($FG ne "") { 
        if ($BG < 1 ) { $BG = "0"; }
        $CODE .= "<SPAN CLASS=C$BG><FONT COLOR=$COLORCODE{$FG}>"; }
      $CHAR=$CODE;
    }
    if ($CHAR eq "") {
      my($CODE)="";
      $CODE .= "</B>" if ($BOLD==1);
      $CODE .= "</U>" if ($UL==1);
      $CODE .= "</FONT></SPAN>" if ($COL==1);
      $CODE .= "</FONT></SPAN>" if ($REV==1);
      $BOLD=0; $UL=0; $COL=0; $REV=0;
      $CHAR=$CODE;
    }
    $NEWMESSAGE .= $CHAR;
    $POSITION++;
  }
  return $NEWMESSAGE;
} 

# Builds our tcp socket with the irc server

sub buildsocket {
 my($HOST, $PORT, $NICK, $REALNAME)=@_;
 my($IDENT);

 $SOCKET = new IO::Socket::INET( Proto   => "tcp",
                                 PeerAddr=> "$HOST:$PORT" );
 if (! defined $SOCKET) { return "Error while connecting to $PORT: $!\n"; }

 my($LOCALPORT) = sockaddr_in($SOCKET->sockname);

 if ($FORM{'IDENT_METHOD'} eq "REMOTE_ADDR") {
  my(@IP)=split(/\./, $ENV{'REMOTE_ADDR'});
  $IDENT = sprintf("%2x%2x%2x%2x", @IP);
  $IDENT =~ s/ /0/g;
  $IDENT =~ tr/a-z/A-Z/;
 }
 else {
  $IDENT = $FORM{'nick'}; 
  $IDENT =~ tr/A-Z/a-z/;
 }

 # Create an entry in the IDENT mapping table
 if ($CONFIG{'USE_IDENT'} == 1) {
  my(%PIDTABLE);
  tie %PIDTABLE, SDBM_File, $CONFIG{'IDENTTABLE'}, O_RDWR | O_CREAT, 0644;
  $PIDTABLE{$LOCALPORT}="$$:$IDENT";
  untie %PIDTABLE;
 }

 select($SOCKET); $| = 1; select(STDOUT);

 my($hostname)=&hostname;
 my($theirhost)= gethostbyaddr(pack('C4', split(/\./, $ENV{'REMOTE_ADDR'})), 2); 

 print $SOCKET  "USER $IDENT $theirhost $hostname :$REALNAME\015\012";
 print $SOCKET  "NICK $NICK\015\012";
 return 1;
}

sub output {
  my($output)=@_;
  chomp($output);
  $output = &colorcode($output);
  print "$output<BR><SCRIPT>$CONFIG{'SCROLL'}</SCRIPT>\015\012";
  return;
}


sub from {
  my($from) = @_;
  my($han, $addy) = split(/\!/, $from);
  $han =~ s/^\://;
  return ($han, $addy);
}

# We need a compare subroutine because using a regex
# could result in an inaccurate result, since people
# can have |, ^, and other special characters in
# their handle.

sub compare {
  my($a, $b) = @_;
  $a =~ tr/A-Z/a-z/;
  $b =~ tr/A-Z/a-z/;
  chomp($a, $b);
  $a =~ s/^\s+//; $a =~ s/\s+$//;
  $b =~ s/^\s+//; $b =~ s/\s+$//;
  if ($a eq $b) { return 1; }
  return 0;
}

sub userlist {

  # Call the appropriate userlist subroutine depending on which browser
  # the client is using
  if ($FORM{'BROWSER'} eq "MSIE") { &userlist_msie(@_); }
  if ($FORM{'BROWSER'} eq "NETSCAPE") { &userlist_ns(@_); }
}

sub userlist_msie {
  my($line, @lines, $COMMAND, @USERS);
  my($MESSAGE)=@_;
  ($MESSAGE)=split(/\002/, $MESSAGE);
  @lines=split(/\n/, $MESSAGE);
  foreach $line (@lines) {
    chomp $line;
    ($COMMAND, @USERS)=split(/ /, $line);

    if ($COMMAND eq "\004REMOVE") {
      foreach(@USERS) {
        next unless ($_);
        my($USERNAME)=$_;
        $USERNAME =~ tr/A-Z/a-z/;
        delete($CURRENT_USERS{$USERNAME});
        delete($CURRENT_MODES{$USERNAME});
        $DEBUG && &output("debug: userlist: removing $USERNAME");
      }
      next;
    }

   if ($COMMAND eq "\004MODE") {
     foreach(@USERS) {
       next unless($_);
       my($USERNAME, $MODE)=split(/\:/);
       $USERNAME =~ tr/A-Z/a-z/;
       $CURRENT_MODES{$USERNAME}=$MODE if ($CURRENT_USERS{$USERNAME});;
       $DEBUG && &output("debug: userlist: mode change $MODE $USERNAME");
     }
   }

   if ($COMMAND eq "\004NICK") {
     foreach(@USERS) {
       next unless($_);
       my($USERNAME, $NEWUSERNAME)=split(/\:/);
       $USERNAME =~ tr/A-Z/a-z/;
       my($NEWNAME) = $NEWUSERNAME;
       $NEWUSERNAME =~ tr/A-Z/a-z/;
       $CURRENT_MODES{$NEWUSERNAME}=$CURRENT_MODES{$USERNAME};
       $CURRENT_USERS{$NEWUSERNAME}=$NEWNAME;
       delete($CURRENT_MODES{$USERNAME});
       delete($CURRENT_USERS{$USERNAME});
       $DEBUG && &output("debug: userlist: nick change $USERNAME -> $NEWNAME");
     }
  }

   if ($COMMAND eq "\004NEW") {
     undef(%CURRENT_USERS);
     undef(%CURRENT_MODES);
     $COMMAND = "\004ADD";
     $DEBUG && &output("debug: userlist: clearing userlist");
   }

   if ($COMMAND eq "\004ADD") {
     foreach(@USERS) {
       $_ =~ s/\002//g;
       $_ =~ s/\004//g;
       next unless ($_);
       my($USERNAME)=$_;
        s/\@|\+//g;
       $USERNAME =~ tr/A-Z/a-z/;
       $USERNAME =~ s/^\+/\002/;
       $USERNAME =~ s/^\@/\001/;
       my($MODEUSER) = $USERNAME;
       $MODEUSER =~ s/\001|\002//g;
       $CURRENT_MODES{$MODEUSER}="";
       if ($USERNAME =~ /\002/) { $CURRENT_MODES{$MODEUSER}="\002"; }
       if ($USERNAME =~ /\001/) { $CURRENT_MODES{$MODEUSER}="\001"; }
       if ($CURRENT_MODES{$MODEUSER} ne "")
         { $USERNAME = $CURRENT_MODES{$MODEUSER} . $MODEUSER; }

       $CURRENT_USERS{$MODEUSER}=$_;
       $DEBUG && output("debug: userlist: adding $MODEUSER");
     }
     next;
    }
  }
  &update_msie;
  return;
}

sub update_msie {
  my($OUTPUT)="";
  $OUTPUT="<CODE>" if ($FORM{'fixed'});
  my(@fields)=keys(%CURRENT_USERS);
  my(%prints);
  foreach(@fields) {
    my($NAME)="";
    if ($CURRENT_MODES{$_} eq "\001") { $NAME .= "@"; }
    if ($CURRENT_MODES{$_} eq "\002") { $NAME .= "+"; }
    $prints{"$CURRENT_MODES{$_}$_"}="$NAME$CURRENT_USERS{$_}";
  }
  @fields=sort(keys(%prints));
  foreach(@fields) { 
    my($name)=$prints{$_};
    $name =~ s/</\&lt\;/g;
    $name =~ s/>/\&gt\;/g;
    $OUTPUT .= "$name<BR>" if ($prints{$_} ne ""); 
  }
  $OUTPUT .= "</CODE>" if ($FORM{'fixed'});
  $OUTPUT =~ s/\n|\013|\000//g;
  if ($DEBUG>1) { 
   print "OUTPUT='$OUTPUT'<br>";
   for(0..length($OUTPUT)) {
    $x = substr($OUTPUT, $_, 1);
    print "CHAR: '$x' ASCII: ".ord($x)."<br>";
   }
   print "\n";
  }
  print "<SCRIPT>parent.set_userlist(\"$OUTPUT\");</SCRIPT>\n";
  return;
}

sub pshmwrite {
  my($KEY, $MESSAGE, $ID, $LENGTH)=@_;
  if ($CONFIG{'USE_SHMEM'}==1) {
    shmwrite($KEY, $MESSAGE, $ID, $LENGTH) || die "shmwrite: $!\n"; }
  else {
    open(MEM, ">$CONFIG{'MEMDIR'}/$KEY") || die "$CONFIG{'MEMDIR'}/$KEY: $!\n";
    print MEM $MESSAGE, "\n";
    close(MEM);
  }
  return;
}

sub pshmread {
  my($KEY, $ID, $LENGTH)=@_;
  my($BUFFER);
  if ($CONFIG{'USE_SHMEM'}==1) 
    { shmread($KEY, $BUFFER, $ID, $LENGTH) || die "shmread: $!\n"; }
  else {
    open(MEM, "<$CONFIG{'MEMDIR'}/$KEY") || die "$CONFIG{'MEMDIR'}/$KEY: $!\n";
    chomp; $BUFFER = <MEM>;
    close(MEM);
  }
  return $BUFFER;
}


sub userlist_ns {
  return unless ($FORM{'USERKEY'});
  my($MESSAGE)=@_;
  my($buff)=&pshmread($FORM{'USERKEY'}, 0, 20);
  my($wait)=0;
  while($buff !~ /^LIST\:/) {
    $wait++;
    $buff = &pshmread($FORM{'USERKEY'}, 0, 20);

# Sends an ALRM message to attempt to regain control
# of the userlist.  Rather than do this I had the
# userlist script reset the memory every 20 seconds.
#
    if ( (int($wait/5) == ($wait/5)) &&
      ($LASTKNOWNPID)) { kill ALRM, $LASTKNOWNPID; }

    $buff="LIST:$LASTKNOWNPID:" if ($wait>15);
    select(undef, undef, undef, .50);
  }
  my($pid);
  ($pid, $pid)=split(/\:/, $buff);
  $LASTKNOWNPID=$pid;
  &pshmwrite($FORM{'USERKEY'}, $MESSAGE, 0, length($MESSAGE));
  kill ALRM, $pid;
}

sub pshmwrite {
  my($KEY, $MESSAGE, $ID, $LENGTH)=@_;
  if ($CONFIG{'USE_SHMEM'}==1) {
    shmwrite($KEY, $MESSAGE, $ID, $LENGTH) || die "shmwrite: $!\n"; }
  else {
    open(MEM, ">$CONFIG{'MEMDIR'}/$KEY") || die "$CONFIG{'MEMDIR'}/$KEY: $!\n";
    print MEM $MESSAGE, "\n";
    close(MEM);
  }
  return;
}

sub pshmread {
  my($KEY, $ID, $LENGTH)=@_;
  my($BUFFER);
  if ($CONFIG{'USE_SHMEM'}==1) 
    { shmread($KEY, $BUFFER, $ID, $LENGTH) || die "shmread: $!\n"; }
  else {
    open(MEM, "<$CONFIG{'MEMDIR'}/$KEY") || die "$CONFIG{'MEMDIR'}/$KEY: $!\n";
    chomp; $BUFFER = <MEM>;
    close(MEM);
  }
  return $BUFFER;
}

sub ReadParse {
 my(%F, $buffer, @pairs);
 if ($ENV{'REQUEST_METHOD'} =~ /GET/i)
  { @pairs = split(/&/, $ENV{'QUERY_STRING'}); }
 else {
  read(STDIN, $buffer, $ENV{'CONTENT_LENGTH'});
  @pairs = split(/\&/, $buffer);
 }
 foreach(@pairs) {
  my($name, $value) = split(/\=/, $_);

  $name =~ tr/+/ /;
  $name =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
  $value =~ tr/+/ /;
  $value =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
  $value =~ s/<!--(.|\n)*-->//g;
  $F{$name}=$value;
 }

 return %F;
}
