#!/usr/local/bin/perl

# Whiplash v1.4
# Copyright (c) 1997-2000, Network Dweebs
# All Rights Reserved

use SDBM_File;
use Fcntl;
use strict vars;
use vars qw { %CONFIG @DISALLOWED_CHANNELS %COLORCODE %FORM %CURRENT_USERS
              %CURRENT_MODES };
require ".config"; 

%FORM = &ReadParse;
my($KEY) = $FORM{'USERKEY'};

# We still import $FORM{'BROWSER'} in the event that a browser other than
# netscape suddenly finds its way onto the market and we need to
# distinguish minor code differences between the two.  OR if MSIE
# ever works right and uses this script.

# Enter our PID and SHMKEY into the PID Table for Leak Cleanup
# Some operating systems do not report any signal upon losing
# the client, hence cleanup never gets called.
do {
 my(%PIDTABLE);
 tie %PIDTABLE, SDBM_File, $CONFIG{'PIDTABLE'}, O_RDWR | O_CREAT, 0644;
 $PIDTABLE{$KEY}=$$;
 untie %PIDTABLE;
};

$CONFIG{'IPC_RMID'} = 0;

# We use a : separator here and a \002 separator in the message because
# When you overwrite shared memory, it only overwrites the length of
# the message, leaving any old junk lying around.

sub readymem {
 my($MESSAGE) = "LIST:$$:";
 &pshmwrite($KEY, $MESSAGE, 0, length($MESSAGE));
}

setpriority 0, 0, getpriority(0, 0) + $CONFIG{'PRIORITY'};

$| = 1;

$SIG{'HUP'}="IGNORE";
$SIG{'INT'}=\&dieout;
$SIG{'ALRM'}=\&msg;     # Alarm from nph-irc.cgi to update the userlist 

sub dieout {
 my(%PIDTABLE);
 if ($CONFIG{'USE_SHMEM'}==1) { shmctl($KEY ,$CONFIG{'IPC_RMID'}, 0); }
 else { unlink("$CONFIG{'MEMDIR'}/$KEY"); }
 tie %PIDTABLE, SDBM_File, $CONFIG{'PIDTABLE'}, O_RDWR | O_CREAT, 0644;
 delete($PIDTABLE{$KEY});
 untie %PIDTABLE;
 exit;
}

# Here we read in the message from our shared memory buffer, 
# interpret it, and update the user list. 

sub msg {
  my($line)=&pshmread($KEY, 0, 4096);
  ($line)=split(/\002/, $line);
  chomp($line);
  my(@lines)=split(/\n/, $line);
  foreach $line (@lines) {
    chomp $line;
    my($COMMAND, @USERS)=split(/ /, $line);
 
    if ($COMMAND eq "\004REMOVE") {
      foreach(@USERS) {
        next unless ($_);
       my($USERNAME)=$_;
       $USERNAME =~ tr/A-Z/a-z/;
       delete($CURRENT_USERS{$USERNAME});
       delete($CURRENT_MODES{$USERNAME});
     }
     next;
   }

   if ($COMMAND eq "\004MODE") {
     foreach(@USERS) {
       next unless($_);
       my($USERNAME, $MODE)=split(/\:/);
       $USERNAME =~ tr/A-Z/a-z/;
       $CURRENT_MODES{$USERNAME}=$MODE if ($CURRENT_USERS{$USERNAME});
     }
   } 

   if ($COMMAND eq "\004NICK") {
     foreach(@USERS) {
       next unless($_);
       my($USERNAME, $NEWUSERNAME)=split(/\:/);
       $USERNAME =~ tr/A-Z/a-z/;
       my($NEWNAME) = $NEWUSERNAME;
       $NEWUSERNAME =~ tr/A-Z/a-z/;
       $CURRENT_MODES{$NEWUSERNAME}=$CURRENT_MODES{$USERNAME};
       $CURRENT_USERS{$NEWUSERNAME}=$NEWNAME;
       delete($CURRENT_MODES{$USERNAME});
       delete($CURRENT_USERS{$USERNAME}); 
     }
  } 
 
   if ($COMMAND eq "\004NEW") {
     undef(%CURRENT_USERS);
     undef(%CURRENT_MODES);
     $COMMAND = "\004ADD";
   }

   if ($COMMAND eq "\004ADD") {
     foreach(@USERS) {
       $_ =~ s/\002//g;
       $_ =~ s/\004//g;
       next unless ($_);
       my($USERNAME)=$_;
        s/\@|\+//g;
       $USERNAME =~ tr/A-Z/a-z/;
       $USERNAME =~ s/^\+/\002/;
       $USERNAME =~ s/^\@/\001/;
       my($MODEUSER) = $USERNAME;
       $MODEUSER =~ s/\001|\002//g;
       $CURRENT_MODES{$MODEUSER}="";
       if ($USERNAME =~ /\002/) { $CURRENT_MODES{$MODEUSER}="\002"; }
       if ($USERNAME =~ /\001/) { $CURRENT_MODES{$MODEUSER}="\001"; }
       if ($CURRENT_MODES{$MODEUSER} ne "")
         { $USERNAME = $CURRENT_MODES{$MODEUSER} . $MODEUSER; }
       
       $CURRENT_USERS{$MODEUSER}=$_;
     }
     next;
    }
  }
  &update;
  &readymem;
  return;
}


# This subroutine runs through the $CURRENT_USERS hash and
# prints out all the current users (simple huh?).  In netscape
# when you reprint the boundary it clears the screen so you can
# reprint text on it (nice).  I have yet to find out how to do this
# in MSIE though.  If you know of a javascript or other alternative
# method for doing this in MSIE, please let me know.

sub update {
  if ($FORM{'BROWSER'} eq "NETSCAPE") {
    print "--BOUNDARY\n";
    print "Content-type: text/html\n\n";
    print "<HTML><BODY BGCOLOR=#C3C3C3 TEXT=#000060>";
  }
  print "<CODE>\n" if ($FORM{'fixed'});
  my(@fields)=keys(%CURRENT_USERS);
  my(%prints);
  foreach(@fields) {
    my($NAME)="";
    if ($CURRENT_MODES{$_} eq "\001") { $NAME .= "@"; }
    if ($CURRENT_MODES{$_} eq "\002") { $NAME .= "+"; }
    $prints{"$CURRENT_MODES{$_}$_"}="$NAME$CURRENT_USERS{$_}";
  }
  @fields=sort(keys(%prints));
  foreach(@fields) { 
    my($name)=$prints{$_};
    $name =~ s/</\&lt\;/g;
    $name =~ s/>/\&gt\;/g;
    print "$name<BR>\n" if ($prints{$_} ne ""); 
  }
  print "<BR><BR><BR><BR>\n" if ($#fields<2);
  print "</CODE>\n" if ($FORM{'fixed'});
  return;
}

# Print the magic header to create a server push, also
# referred to as a 'server stream'.  Great for chatrooms *g*
 
print "HTTP/1.0 200\n";
print "Content-type: multipart/x-mixed-replace;boundary=BOUNDARY\n\n";

if ($FORM{'BROWSER'} eq "NETSCAPE") {
  print "--BOUNDARY\n";
  print "Content-type: text/html\n\n";
  print "<HTML><BODY BGCOLOR=#FFFFFF TEXT=#000060>";
}
print "<B><FONT COLOR=#000060>Initializing...</B><BR>\n" .

&readymem;
print "<NOBR>RAM Allocated.<BR>\n";

my($PARENT)=&getproc;
if ($PARENT == $FORM{'PARENTPID'}) { $PARENT=0; }
my($COUNT)=0;
while(!($PARENT)) {
  select(undef, undef, undef, .25);
  $PARENT=&getproc;
  if ($PARENT == $FORM{'PARENTPID'}) { $PARENT=0; }
  $COUNT++;
  if ($COUNT>20) { die "Couldn't find Parent PID\n"; }
}

print "Found SocketPID.<BR>\n";
print "Waiting...<BR></NOBR>\n";

while(kill(0, $PARENT)) {
  $COUNT=0;
  select(undef, undef, undef, .25);
  &readymem if (int($COUNT/40)==($COUNT/40));
  $COUNT=0 if ($COUNT>1000);
}

exit;


sub getproc {
 my(%PIDTABLE);
 tie %PIDTABLE, SDBM_File, $CONFIG{'PIDTABLE'}, O_RD, 0644;
 my($proc)=$PIDTABLE{$FORM{'IPCKEY'}};
 untie %PIDTABLE;
 return $proc;
}

sub pshmwrite {
  my($KEY, $MESSAGE, $ID, $LENGTH)=@_;
  if ($CONFIG{'USE_SHMEM'}==1) {
    shmwrite($KEY, $MESSAGE, $ID, $LENGTH) || die "shmwrite: $!\n"; }
  else {
    open(MEM, ">$CONFIG{'MEMDIR'}/$KEY");
    print MEM $MESSAGE, "\n";
    close(MEM);
  }
  return;
}

sub pshmread {
  my($KEY, $ID, $LENGTH)=@_;
  my($BUFFER);
  if ($CONFIG{'USE_SHMEM'}==1)
    { shmread($KEY, $BUFFER, $ID, $LENGTH) || die "shmread: $!\n"; }
  else {
    open(MEM, "<$CONFIG{'MEMDIR'}/$KEY") || die "$CONFIG{'MEMDIR'}/$KEY: $!\n";
    chomp; $BUFFER = <MEM>;
    close(MEM);
  }
  return $BUFFER;
}

sub ReadParse {
 my(%F, $buffer, @pairs);
 if ($ENV{'REQUEST_METHOD'} =~ /GET/i)
  { @pairs = split(/&/, $ENV{'QUERY_STRING'}); }
 else {
  read(STDIN, $buffer, $ENV{'CONTENT_LENGTH'});
  @pairs = split(/\&/, $buffer);
 }
 foreach(@pairs) {
  my($name, $value) = split(/\=/, $_);

  $name =~ tr/+/ /;
  $name =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
  $value =~ tr/+/ /;
  $value =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
  $value =~ s/<!--(.|\n)*-->//g;
  $F{$name}=$value;
 }

 return %F;
}
